/*!
 * @header CA-Lava Data
 * Convenience routines to push data around between the tasks.
 */

#ifndef __CA_LAVA_DATA__
#define __CA_LAVA_DATA__

#include "v2int.h"

/*!
 * @typedef worker_info
 * Data describing the position of a worker thread in the board setup.
 * @field pos upper left corner of the worker's cell in absolute board
 * coordinates
 * @field next array of adjacent workers
 */
typedef struct worker_info {
  v2int org;
  int next[4];
} worker_info;

/*!
 * @typedef cell
 * All the status data of a single cell on the simulation board.
 * @field z height of cell floor above mean sea level
 * @field h thickness of lava layer
 * @field T mean temperature of lava layer
 * @field i thickness of lava flowing in from the source every step
 * @field U temperature of lava flowing in from the source every step
 * @field o thickness of lava flowing out of the cell in the four main
 * geographical directions
 * @field tmp temporary state variable during transformations
 */
typedef struct cell {
  float z;
  float h, T;
  float i, U;
  float o[4];
  float tmp;
} cell;

/*!
 * @function send_config
 * Sends all the necessary configuration from the master to a worker.
 * @param tid task id of receiver
 * @param info worker_info structure describing the position of the
 * receiver in the board layout
 * @param cell_dim cell dimensions for the worker
 * @param step_count number of steps to simulate
 * @param step_interval number of steps between data dumps to the master
 * @param solidus lava solidus temeperature
 * @param adhesion lava adhesion coefficients
 * @param cooling lava cooling coefficient
 */
void send_config(int tid, const worker_info *info, v2int cell_dim, int step_count, int step_interval, float solidus, const float adhesion[], float cooling);

/*!
 * @function recv_config
 * Receives all the necessary configuration for a worker from the master.
 * @param tid task id of master
 * @param org position of the receiver in the board layout
 * @param next array of task ids of the neighbouring workers
 * @param cell_dim cell dimensions for the worker
 * @param step_count number of steps to simulate
 * @param step_interval number of steps between data dumps to the master
 * @param solidus lava solidus temeperature
 * @param adhesion lava adhesion coefficients
 * @param cooling lava cooling coefficient
 */
void recv_config(int tid, v2int *org, int next[], v2int *cell_dim, int *step_count, int *step_interval, float *solidus, float adhesion[], float *cooling);

/*!
 * @function send_board
 * Sends a rectangular subsection of the full gameboard or a simulation
 * block to another task.
 * @param tid task ID of recipient
 * @param board pointer to board data array
 * @param org origin of the simulation block in absolute full board
 * coordinates. Set to V2INT(0, 0) if you are transmitting from the
 * full board.
 * @param dim dimensions of the board or simulation block
 * @param pos position of upper left corner to be transmitted in board
 * coordinates relative to the origin of the simulation block
 * @param size dimensions of block to send
 */
void send_board(int tid, const cell board[], v2int org, v2int dim, v2int pos, v2int size);

/*!
 * @function recv_board
 * Receives a rectangular subsection of the full gameboard or a
 * simulation block from another task.
 * @param tid task ID of sender
 * @param board pointer to board data array where the data will be
 * incorporated
 * @param org origin of the simulation block in absolute full board
 * coordinates. Set to V2INT(0, 0) if you are receiving into the
 * full board.
 * @param dim dimensions of the board or simulation block
 */
void recv_board(int tid, cell board[], v2int org, v2int dim);

#endif /* __CA_LAVA_DATA__ */

/* vim:set shiftwidth=2 softtabstop=2: */
