/*!
 * @header CA-Lava Transform
 * The transformation for a simulation step of the lava flows.
 */

#ifndef __CA_LAVA_TRANSFORM__
#define __CA_LAVA_TRANSFORM__

#include <math.h>
#include <stdio.h>

#include "ca-lava-data.h"
#include "config.h"

/*!
 * @var solidus
 * Lava solidus temperature
 */
extern float solidus;

/*!
 * @var adhesion
 * Lava adhesion coefficients
 */
extern float adhesion[2];

/*!
 * @var cooling
 * Lava cooling coefficient
 */
extern float cooling;

/*!
 * @define FILL
 * Filling value used for "empty" variables.
 */
#define FILL (-9999.f)


/*
 * @function tf_mix
 * First board state tranformation step. Mixes lava that stays in the
 * cell with that coming in from sources and neighbours. Calculates the
 * mixture temperature and new thickness of lava layer after in- and
 * outfluxes have been applied.
 * @param this the cell being transformed
 * @param next array of neighbour cells
 */
#ifndef INSTANTIATE
static inline
#endif
void tf_mix(cell *this, const cell *next[]) {
  int j;

  /* this cell */
  this->tmp = this->h * this->T;
  this->tmp += this->i * this->U;
  this->h += this->i;

  /* neighbours */
  for (j = 0; j < 4; j++) {
    float i =  next[j]->o[(j + 2) % 4];
    this->tmp += next[j]->T * i;
    this->h += i;
  }

  /* average */
  if (this->h > 0.f)
    this->tmp /= this->h;
  else
    this->tmp = FILL;

  /* outflux */
  for (j = 0; j < 4; j++)
    this->h -= this->o[j];
}

/*
 * @function tf_cool
 * Second board state transformation step. Calculates the effect of
 * radiative lava cooling. Solidifies the lava if necessary.
 * @param this the cell being transformed
 * @param next array of neighbour cells
 */
#ifndef INSTANTIATE
static inline
#endif
void tf_cool(cell *this, const cell *next[]) {
  if (this->h > 0.f) {
    /* temperature after radiative cooling */
    this->T =
      this->tmp
      / powf(1.f + this->tmp*this->tmp*this->tmp * cooling / this->h, 1.5f);
    
    /* solidify the lava if below solidus and seal the vent */
    if (this->T <= solidus) {
      this->z += this->h;
      this->h = 0.f;
      this->i = 0.f;
    }
  }
}

/*
 * @function tf_equalize
 * Third board state transformation step. Calculates average heights of
 * lavas in adjacent cells and determines lava outflows accordingly.
 * @param this the cell being transformed
 * @param next array of neighbour cells
 */
#ifndef INSTANTIATE
static inline
#endif
void tf_equalize(cell *this, const cell *next[]) {
  int j, loop = 1, active[4] = { 1, 1, 1, 1 };
  float viscous, base, avg, mine, yours[4];

  /* determine base level = absolute height where retained lava ends */
  viscous = adhesion[0] * expf(-adhesion[1] * this->T);
  base = this->z + viscous;
  /* determine offset of our lava height over base level */
  mine = this->h - viscous;

  /* initialize all outflows to zero */
  for (j = 0; j < 4; j++)
    this->o[j] = 0.f;

  /* only proceed if there is any lava that can flow out */
  if (mine > 0.f) {
    /* determine offsets of the adjacent lava levels above base level */
    for (j = 0; j < 4; j++)
      yours[j] = next[j]->z + next[j]->h - base;
    
    /* determine average level and neighbour cell below it */
    while (loop) {
      int count;
      
      loop = 0;
      avg = mine;
      
      for (j = 0, count = 0; j < 4; j++)
	if (active[j]) {
	  avg += yours[j];
	  count++;
	}
      
      avg /= count + 1;
      
      for (j = 0; j < 4; j++)
	if (active[j] && (yours[j] > avg)) {
	  active[j] = 0;
	  loop = 1;
	}
    }

    /* only proceed pouring something out if there is lava above average */
    if (mine > avg) {
      /* if the average level is above base level, distribution is easy */
      if (avg > 0.f) {
	for (j = 0; j < 4; j++) {
	  if (active[j])
	    this->o[j] = avg - yours[j];
	}
      }
      /* otherwise we must avoid pouring out solid rock or retained lava */
      else {
	base = 0.f;
	for (j = 0; j < 4; j++)
	  if (active[j]) {
	    yours[j] = avg - yours[j];
	    base += yours[j];
	  }

	mine /= base;

	for (j = 0; j < 4; j++) {
	  if (active[j])
	    this->o[j] = mine * yours[j];
	}
      }
    }
  }
}

#endif /* __CA_LAVA_TRANSFORM__ */

/* vim:set shiftwidth=2 softtabstop=2: */
