/*!
 * @header CA-Lava Worker
 * Worker task for a distributed simulation of lava flows by cellular
 * automata. Spawned by the master and supplied with configuration
 * information as well as initial state. Simulates generation steps of
 * the board, synchronizes the simulation block's boundaries with
 * workers handling the neighbouring blocks and periodically send
 * results back to the master.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <pvm3.h>

#include "v2int.h"
#include "exn.h"
#include "ca-lava-data.h"
#include "ca-lava-transform.h"
#include "utils.h"
#include "config.h"

/*!
 * @function sync_boards
 * Function to synchronize board states between the workers
 * @param next array of neighbour task ids
 * @param org origin of own board region
 * @param dim dimension of the board
 * @param dim_2 reduced dimension of the board (excluding rim)
 * @param board board cell array
 */
static inline void sync_boards(int next[], const v2int org, const v2int dim, const v2int dim_2, cell *board) {
  int j;

  /*
   * Transmit states in the new board's inner boundary to adjacent
   * workers
   */
  /* north */
  if (next[DIR_NORTH])
    send_board(next[DIR_NORTH], board, org, dim,
	       V2INT(1, 1), V2INT(dim_2.x, 1));
  /* east */
  if (next[DIR_EAST])
    send_board(next[DIR_EAST], board, org, dim,
	       V2INT(dim_2.x, 1), V2INT(1, dim_2.y));
  /* south */
  if (next[DIR_SOUTH])
    send_board(next[DIR_SOUTH], board, org, dim,
	       V2INT(1, dim_2.y), V2INT(dim_2.x, 1));
  /* west */
  if (next[DIR_WEST])
    send_board(next[DIR_WEST], board, org, dim,
	       V2INT(1, 1), V2INT(1, dim_2.y));
  
  /*
   * And pull in the changes sent by the adjacent workers
   */
  for (j = 0; j < 4; j++) {
    if (next[j])
      recv_board(next[j], board, org, dim);
  }
}

/*!
 * @function main
 * Entry point for the worker tasks. Doesn't take any command line
 * arguments but receives necessery information from the master which
 * spawned it.
 */
int main(void) {
  int i;
  cell *board = NULL;
  const cell **topo = NULL;
  EXN_DURING {
    int tid, parent;
    v2int org, dim, dim_2;
    int next[4];
    int step_count, step_interval;
    v2int pos;

    if ((tid = pvm_mytid()) < 0)
      EXN_THROW(EXN_CODE_PVM, "couldn't get my TID");
    if ((parent = pvm_parent()) < 0)
      EXN_THROW(EXN_CODE_PVM, "couldn't get parent's TID");

    printf("CA-Lava Worker t%x.t%x coming up\n", parent, tid);

    /*
     * Receive initialization information from the master task
     */
    recv_config(parent, &org, next, &dim,
		&step_count, &step_interval,
		&solidus, adhesion, &cooling);

    printf("org = (%d, %d), dim = (%d, %d), next = (t%x, t%x, t%x, t%x)\n"
	   "step_count = %d, step_interval = %d\n"
	   "solidus = %g, adhesion = (%g, %g), cooling = %g\n",
	   org.x, org.y, dim.x, dim.y,
	   next[DIR_NORTH], next[DIR_EAST], next[DIR_SOUTH], next[DIR_WEST],
	   step_count, step_interval,
	   solidus, adhesion[0], adhesion[1], cooling);

    /*
     * Allocate board and receive initial state from the master task
     */
    if (!(board = NEWARRAY(cell, dim.x * dim.y)))
      EXN_THROW(EXN_CODE_STDLIB, "couldn't allocate board");

    if (!(topo = NEWARRAY(const cell *, 4 * dim.x * dim.y)))
      EXN_THROW(EXN_CODE_STDLIB, "couldn't allocate board topology cache");

    puts("receiving board");
    recv_board(parent, board, org, dim);

    /*
     * Create a board topology cache
     */
    for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
      for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	int ofs = V2INT_OFS(pos, dim), ofs4 = 4 * ofs;
	topo[ofs4 + DIR_NORTH] = &board[ofs - dim.x];
	topo[ofs4 + DIR_EAST] = &board[ofs + 1];
	topo[ofs4 + DIR_SOUTH] = &board[ofs + dim.x];
	topo[ofs4 + DIR_WEST] = &board[ofs - 1];
      }
    
    puts("simulation begins");

    /*
     * For every step, walk the board and apply the transformation rules
     */
    dim_2 = dim;
    V2INT_SUB(dim_2, V2INT(2, 2));

    for (i = 1; i <= step_count; i++) {
      int j;

      /* lava mixing step */
      for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
	for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	  j = V2INT_OFS(pos, dim);
	  tf_mix(&board[j], &topo[4 * j]);
	}

      /* cooling step */
      for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
	for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	  j = V2INT_OFS(pos, dim);
	  tf_cool(&board[j], &topo[4 * j]);
	}

      /* our own height and elevation may have changed, so we resynchronize */
      sync_boards(next, org, dim, dim_2, board);

      /* outflow recalculation step */
      for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
	for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	  j = V2INT_OFS(pos, dim);
	  tf_equalize(&board[j], &topo[4 * j]);
	}

#ifndef NDEBUG
      /* bogus value check */
      for (pos.y = 1; pos.y < dim.y - 1; pos.y++)
	for (pos.x = 1; pos.x < dim.x - 1; pos.x++) {
	  cell *this = &board[V2INT_OFS(pos, dim)];

	  if (this->h > 0.f) {
	    if ((this->T < 0.f) || isnan(this->T) || isinf(this->T))
	      fprintf(stderr,
		      "%s:%d: warning: step %d, cell (%d, %d):"
		      " bogus temperature = %g\n",
		      __FILE__, __LINE__, i, pos.x, pos.y, this->T);
	  }
	  else if ((this->h < 0.f) || isnan(this->h) || isinf(this->h))
	    fprintf(stderr,
		    "%s:%d: warning: step %d, cell (%d, %d):"
		    " bogus height = %g\n",
		    __FILE__, __LINE__, i, pos.x, pos.y, this->h);

	  for (j = 0; j < 4; j++) {
	    if ((this->o[j] < 0.f) || isnan(this->o[j]) || isinf(this->o[j]))
	      fprintf(stderr,
		      "%s:%d: warning: step %d, cell (%d, %d):"
		      " bogus outflux[%d] = %g\n",
		      __FILE__, __LINE__, i, pos.x, pos.y, j, this->o[j]);
	  }
	}
#endif /* NDEBUG */

      sync_boards(next, org, dim, dim_2, board);

      /*
       * Send data back to the master if desired
       */
      if ((i % step_interval == 0) || (i == step_count)) {
	printf("t%x.t%x sending generation %d to master\n", parent, tid, i);
	send_board(parent, board, org, dim, V2INT(1, 1), dim_2);
      }
    }

    puts("simulation complete");
  }
  EXN_ALWAYS {
    if (board) free(board);
    if (topo) free(topo);
  }
  EXN_HANDLER {
    fprintf(stderr, "%s:%d: ", exn_current.file, exn_current.line);
    switch (exn_current.code) {
    case EXN_CODE_STDLIB:
      perror(exn_current.msg);
      break;
    case EXN_CODE_PVM:
      pvm_perror((char *)exn_current.msg);
      break;
    default:
      fputs(exn_current.msg, stderr);
      putc('\n', stderr);
    }
  }
  EXN_END

  return (pvm_exit() < 0 ? 1 : 0);
}

/* vim:set shiftwidth=2 softtabstop=2: */
